import tiktoken


def count_messages_tokens(messages: list):
    return sum([count_message_tokens(message) for message in messages])

def count_message_tokens(message: dict):
    
    if "content" in message:
        return count_string_tokens(message["content"])
    else:
        return 0

def count_string_tokens(message: str):
    
    model = "gpt-4"

    encoding = tiktoken.encoding_for_model(model)
    token_count = 4
    token_count += len(encoding.encode(message))
    return int(token_count)



def convert_messages_to_format(messages, provider_name):
    # This function will convert the messages to the format that the provider requires.
    # The provider_name is the name of the provider that the messages are being converted to.
    # The messages are the messages that are being converted.
    # The return value is the messages in the format that the provider requires.
    
    valid_providers = ["openai", "anthropic_messages", "anthropic_completions", "llama", "cohere"]

    if provider_name not in valid_providers:
        raise ValueError(f"Invalid provider name {provider_name}")
    
    if provider_name == "openai":
        return messages
    
    elif provider_name == "anthropic_messages":
        
        if messages[0]["role"] == "system":
            return {"messages": messages[1:], "system": messages[0]["content"]}
        else:
            return messages
        
    elif provider_name == "anthropic_completions":
        human_prefix = "\n\nHuman: "
        assistant_prefix = "\n\nAssistant: "
        msg = _messages_to_string_with_prefixes(messages, human_prefix, assistant_prefix)
        
        return msg
        
    elif provider_name == "cohere":
        human_prefix = "\nHuman: "
        assistant_prefix = "\nAssistant: "
        msg = _messages_to_string_with_prefixes(messages, human_prefix, assistant_prefix)
        
        return msg
    
    elif provider_name == "llama":
        B_INST, E_INST = "[INST]", "[/INST]"
        B_SYS, E_SYS = "<<SYS>>\n", "\n<</SYS>>\n\n"

        # The message roles need to alternate between user and assistant. Optional: start with a system message.
        if messages[0]["role"] == "system":
            messages[1] = {"role": messages[1]["content"], "content": B_SYS + messages[0]["content"] + E_SYS + messages[1]["content"]}
            messages = messages[1:]
        
        if messages[0]["role"] != "user":
            raise ValueError("First message role for Llama provider must be user")
        if messages[1]["role"] != "assistant":
            raise ValueError("Second message role for Llama provider must be assistant")
        if messages[-1]["role"] != "user":
            raise ValueError("Last message role for Llama provider must be user")

        for i in range(1, len(messages)):
            if messages[i]["role"] == messages[i-1]["role"]:
                raise ValueError("Message roles for Llama provider must alternate between user and assistant")
            
        for message in messages:
            message["content"] = message["content"].strip()

        human_prefix = B_INST
        assistant_prefix = E_INST
        msg = _messages_to_string_with_prefixes(messages, human_prefix, assistant_prefix)
        msg = msg + E_INST

        return msg
    
    else:
        raise ValueError(f"Invalid provider name {provider_name}")

def _messages_to_string_with_prefixes(messages, human_prefix, assistant_prefix):
    msg = ""

    for message in messages:
        if message["role"] == "user":
            msg += human_prefix + message["content"]
        elif message["role"] == "assistant":
            msg += assistant_prefix + message["content"]
        elif message["role"] == "system":
            msg += message["content"]
        else:
            raise ValueError(f"Invalid message role {message['role']}")
        
    msg += assistant_prefix
    return msg


from .prompt_template import PromptTemplate
from .database_query_template_format import read_and_format_query
from .utilities import read_prompt_from_system_and_id

class LLMSystem:
    def __init__(self, system_id, systems_location=None):
        self.system_id = system_id
        self.systems_location = systems_location if systems_location is not None else "PROMPT_SYSTEMS_LOCATION"
        if self.systems_location is None:
            raise ValueError("PROMPT_SYSTEMS_LOCATION environment variable must be set or passed as initialization arg.")
        
    def get_formatted_prompt(self, prompt_id: str, variables: dict) -> PromptTemplate:
        prompt = read_prompt_from_system_and_id(self.system_id, self.systems_location, prompt_id)
        return PromptTemplate(prompt, variables)

    def get_formatted_query(self, query_id: str, variables: dict) -> str:
        return read_and_format_query(self.system_id, self.systems_location, query_id, variables)



import json
from .llm_system import LLMSystem
from .prompt_template import PromptTemplate

class SystemLogger:
    def __init__(self, system : LLMSystem, log_location=None) -> None:
        self.system = system
        if log_location is None or isinstance(log_location, str) or hasattr(log_location, 'write'):
            self.log_location = log_location
        else:
            raise ValueError("log_location must be a string or a file pointer or None")

    def log_llm_interaction(self, prompt=None, response=None, prompt_id:str=None, prompt_template:PromptTemplate=None, model:str=None, variables:dict=None):
        # Format a json body containing each added variable plus the system.system_id
        log_body = dict()
        log_body["system_id"] = self.system.system_id
        if prompt is not None:
            log_body["prompt"] = prompt
        elif prompt_template is not None:
            log_body["prompt"] = prompt_template.to_openai()
        elif prompt_id is not None and variables is not None:
            log_body["prompt"] = self.system.get_formatted_prompt(prompt_id, variables).to_openai()

        if response is not None:
            log_body["response"] = response
        if model is not None:
            log_body["model"] = model
        if variables is not None:
            log_body["variables"] = variables

        # If log_location is None, use stdout
        self._log_body(log_body)

    def _log_body(self, log_body):
        if self.log_location is None:
            print(json.dumps(log_body, indent=2))
        else:
            # if log_location is a string open the file and append the log_body, if it is a file pointer write to the file
            if isinstance(self.log_location, str):
                with open(self.log_location, 'a') as f:
                    f.write(json.dumps(log_body))
                    f.write("\n")
            elif hasattr(self.log_location, 'write'):
                self.log_location.write(json.dumps(log_body))
                self.log_location.write("\n")
            else:
                raise ValueError("log_location must be a string or a file pointer")
            
    def log_database_interaction(self, query=None, response=None, query_id:str=None, variables:dict=None):
        # Format a json body containing each added variable plus the system.system_id
        log_body = dict()
        log_body["system_id"] = self.system.system_id
        if query is not None:
            log_body["query"] = query
        elif query_id is not None and variables is not None:
            log_body["query"] = self.system.get_formatted_query(query_id, variables)

        if response is not None:
            log_body["response"] = response
        if variables is not None:
            log_body["variables"] = variables

        # If log_location is None, use stdout
        self._log_body(log_body)


from .utilities import format_template_string, read_query_from_system_and_id

# FreewayAI is property of DiligentlyAI, Inc. and is used under an MIT license.
# by Ben Parfitt, Feb 2024

def _format_query(query, variables):
    query_template = query["search_query_template"]
    return format_template_string(query, query_template, variables)
    

def read_and_format_query(system_id, systems_location, query_id, variables):
    prompt = read_query_from_system_and_id(system_id, systems_location, query_id)
    return _format_query(prompt, variables)


import copy, json
from .prompt_messages_to_provider_format import convert_messages_to_format
from .token_counting import count_messages_tokens
from .utilities import format_messages, format_template_string

class PromptTemplate:
    def __init__(self, template: dict, variables: dict):
        self.template = template
        self.variables = variables
        self.messages = []


        self.added_context = []
        self._format_prompt()

    def _format_prompt(self):

        messages = []
        if "messages" in self.template:
            messages = format_messages(self.template, self.template["messages"], self.variables)
        if "prompt_template" in self.template:
            msg = {
                "content": format_template_string(self.template, self.template["prompt_template"], self.variables),
                "role": "user"
            }
            
            messages.append(msg)
            
        self.messages = messages

        if len(self.added_context) > 0:
            for context in self.added_context:
                self._add_context(context)

    def __str__(self):
        if len(self.messages) == 1:
            return self.messages[0]["content"]
        else:
            result = ""
            for message in self.messages:
                result += f"{message['role']}: {message['content']}\n"
            return result
    
    def __list__(self):
        return self.messages
    
    def count_tokens(self):
        return count_messages_tokens(self.messages)
    
    def trim_tokens(self, max_tokens):
        last_count = self.count_tokens()
        while self.count_tokens() > max_tokens:
            for k, v in self.variables.items():
                
                if len(v) < 100:
                    continue
                else:
                    self.variables[k] = v[int(len(v)*0.1):]
            
            self._format_prompt()

            if self.count_tokens() == last_count:
                raise ValueError(f"Could not trim tokens further (current token count: {self.count_tokens()}; max tokens: {max_tokens})")
            last_count = self.count_tokens()

        return self

    def add_context(self, context):

        # Check the type of context for either string or list
        if type(context) == str:
            context = [{
                "content": context,
                "role": "user"
            }]
            
        if type(context) != list:
            raise ValueError("Context must be either a string or list of message objects in OpenAI format")

        self.added_context.append(context)
        return self._add_context(context)

    def _add_context(self, context):
        if len(self.messages) == 0:
            self.messages = context
        elif self.messages[0]["role"] == "system":
            self.messages = self.messages[:1] + context + self.messages[1:]
        else:
            self.messages = context + self.messages

        return self

    def to_openai(self):
        return convert_messages_to_format(self.messages, "openai")
    
    def to_anthropic_messages(self):
        return convert_messages_to_format(self.messages, "anthropic_messages")
    
    def to_anthropic_completions(self):
        return convert_messages_to_format(self.messages, "anthropic_completions")
    
    def to_llama(self):
        return convert_messages_to_format(self.messages, "llama")
    
    def to_cohere(self):    
        return convert_messages_to_format(self.messages, "cohere")
    



import json

def format_messages(prompt, messages, variables):

    formatted_messages = []

    for message in messages:
        formatted_messages.append({
            "role": message["role"],
            "content": format_template_string(prompt, message["content"], variables)
        })

    return formatted_messages

def read_prompt_from_system_and_id(system_id, systems_location, prompt_id):
    with open(f'{systems_location}/{system_id}/prompt_templates.json') as f:
        prompts = json.load(f)
        target_prompt = prompts["templates"][prompt_id]
        return target_prompt
    
def read_query_from_system_and_id(system_id, systems_location, query_id):
    with open(f'{systems_location}/{system_id}/database_templates.json') as f:
        prompts = json.load(f)
        target_prompt = prompts["templates"][query_id]
        return target_prompt

def format_template_string(body, message_template, variables):
    needed_keys = [entry["name"] for entry in body["variables"]]
    for key in needed_keys:
        if key not in variables:
            raise ValueError(f"Missing required key {key}")
    
    return message_template.format(**variables)